#!/usr/bin/env python3.5 

'''
This work is licensed under the Creative Commons Attribution-ShareAlike 4.0 International License. To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/4.0/.

Website: http://projects.cl.uni-heidelberg.de/SR3de/
Contact: Eva Mujdricza-Maydt, mujdricza@cl.uni-heidelberg.de

'''

from readers import utils_conversion as utils
from readers import CoNLL2009reader


def write_file_with_extracted_annotation(output_file_name, fully_annotated_sentence_list):
    f = open(output_file_name, "w")
    for fully_annotated_sentence in fully_annotated_sentence_list:
        for fully_annotated_line in fully_annotated_sentence:
            f.write(utils.FORMAT.TAB.value.join(fully_annotated_line))
            f.write(utils.FORMAT.NL.value)
        f.write(utils.FORMAT.NL.value)
    f.close()

def get_fully_annotated_VN_sentence_list(srlanno_sr3devn, parallel_entries_for_dataset_dict, conll2009_data_dict):
    fully_annotated_sentence_list = []
    entry_idx = 0 #we get only the first one for this part of the extraction, since all annotations for a sentence are already aggregated
    for sentence_idx in range(len(srlanno_sr3devn)): 
        
        sr3devn_sentence_only_srl = srlanno_sr3devn[sentence_idx]
        
        #we only extract the sentences with at least one annotation
        if sentence_idx in parallel_entries_for_dataset_dict:
            #print("parallel_entries_for_dataset_dict[sentence_idx]: ", parallel_entries_for_dataset_dict[sentence_idx])
            conll_dataset = parallel_entries_for_dataset_dict[sentence_idx][entry_idx][utils.MAPPING_FILE.CoNLL2009_dataset.value]
            conll_sentence_idx = parallel_entries_for_dataset_dict[sentence_idx][entry_idx][utils.MAPPING_FILE.CoNLL2009_sentidx.value]
            conll2009_sentence = conll2009_data_dict[conll_dataset][conll_sentence_idx]
    
            fully_annotated_VN_sentence = __merge_PBannobase_with_VNonlySR(conll2009_sentence, sr3devn_sentence_only_srl)
            fully_annotated_sentence_list.append(fully_annotated_VN_sentence)
        else:
            continue
            
    return fully_annotated_sentence_list

def __merge_PBannobase_with_VNonlySR(conll2009_sentence, sr3devn_sentence_only_srl):
    fully_annotated_sentence = []
    conll2009_sentence_without_srl = __cut_off_srl_annotation(conll2009_sentence)
    utils.check_list_lengths(sr3devn_sentence_only_srl, conll2009_sentence_without_srl)
    
    for i in range(len(conll2009_sentence_without_srl)):
        line_base = conll2009_sentence_without_srl[i]
        line_srl = sr3devn_sentence_only_srl[i]
        
        line = line_base + line_srl
        fully_annotated_sentence.append(line)
    return fully_annotated_sentence

def get_fully_annotated_PB_sentence_list(fully_annotated_VN_sentence_list, 
    parallel_VN2entries_for_dataset_dict, conll2009_data_dict):
    
    #check lengths
    utils.check_list_lengths(fully_annotated_VN_sentence_list, list(parallel_VN2entries_for_dataset_dict.keys()))
    
    fully_annotated_PB_sentence_list = []
        
    # we only extract the sentences with at least one annotation
    sorted_sentence_idxs = sorted(list(parallel_VN2entries_for_dataset_dict.keys()))
    for full_idx in range(len(sorted_sentence_idxs)): 
        sentence_idx = sorted_sentence_idxs[full_idx]
        #get the corresponding sentence in the fully annotated VN sentence list
        srl3devn_sentence = fully_annotated_VN_sentence_list[full_idx]
        
        entries_for_sentence = parallel_VN2entries_for_dataset_dict[sentence_idx]
        entry_idx = 0 #since we already have the full annotation in the srl3devn_sentence
    
        conll_dataset = entries_for_sentence[entry_idx][utils.MAPPING_FILE.CoNLL2009_dataset.value]
        conll_sentence_idx = entries_for_sentence[entry_idx][utils.MAPPING_FILE.CoNLL2009_sentidx.value]
        conll2009_sentence = conll2009_data_dict[conll_dataset][conll_sentence_idx]
        
        #check
        utils.check_list_lengths(srl3devn_sentence, conll2009_sentence)
        
        # take the PB annotation only if it is also annotated in the VN data
        #1) take over predicate annotation
        #2) take over argument annotations
        predicate_position_info = []
        annotated_PB_sentence = []
        argument_column_matrix = []
        for tok_idx in range(len(conll2009_sentence)):
            curr_conll_line = conll2009_sentence[tok_idx]
            curr_conll_line_base = curr_conll_line[0:CoNLL2009reader.CONLL2009_ANNOIDX.FILLPRED.value]
            
            #go over the tokens
            #if there is a FILLEDPRED in conll
            # and also in sr3de --> take this
            # if no sr3de correspondence --> mark the predicate_position for deleting/ignoring
            
            annotated_PB_line_stem = curr_conll_line_base
            
            if curr_conll_line[CoNLL2009reader.CONLL2009_ANNOIDX.FILLPRED.value] == CoNLL2009reader.FORMAT.FILLEDPRED.value:
                if srl3devn_sentence[tok_idx][CoNLL2009reader.CONLL2009_ANNOIDX.FILLPRED.value] == CoNLL2009reader.FORMAT.FILLEDPRED.value:
                    #take
                    predicate_position_info.append(True)
                    annotated_PB_line_stem.append(curr_conll_line[CoNLL2009reader.CONLL2009_ANNOIDX.FILLPRED.value])
                    annotated_PB_line_stem.append(curr_conll_line[CoNLL2009reader.CONLL2009_ANNOIDX.PRED.value])
            
                else:
                    predicate_position_info.append(False)
                    annotated_PB_line_stem.append(CoNLL2009reader.FORMAT.NOVALUE.value)
                    annotated_PB_line_stem.append(CoNLL2009reader.FORMAT.NOVALUE.value)
            else:
                annotated_PB_line_stem.append(CoNLL2009reader.FORMAT.NOVALUE.value)
                annotated_PB_line_stem.append(CoNLL2009reader.FORMAT.NOVALUE.value)
            
            #now, only the argument columns are missing
            annotated_PB_sentence.append(annotated_PB_line_stem)
            
            #also collect the argument columns
            curr_conll_args = curr_conll_line[CoNLL2009reader.CONLL2009_ANNOIDX.APRED1.value:]
            argument_column_matrix.append(curr_conll_args)
        
        utils.check_list_lengths([x for x in predicate_position_info if x == True], entries_for_sentence)
        utils.check_list_lengths(predicate_position_info, argument_column_matrix[0])
        
        #copy the argument annotation if predicate annotation there
        for predicate_position_idx in range(len(predicate_position_info)):
            if predicate_position_info[predicate_position_idx] == True:
                for i in range(len(annotated_PB_sentence)):
                    annotated_PB_sentence[i].append(argument_column_matrix[i][predicate_position_idx])
            else:
                pass
        fully_annotated_PB_sentence_list.append(annotated_PB_sentence)
    
    return fully_annotated_PB_sentence_list

#fully_annotated_FN_sentence_list = converter.get_fully_annotated_FN_sentence_list(sentenceid_to_frames_dict, parallel_VN2entries_for_dataset_dict, fully_annotated_VN_sentence_list)
def get_fully_annotated_FN_sentence_list(salsa_sentenceid_to_frames_dict, parallel_VN2entries_for_dataset_dict, fully_annotated_VN_sentence_list):
    
    fully_annotated_FN_sentence_list = []
    
    entry_index = 0
    sentence_idx_in_vn_list = 0
    # we only extract the sentences with at least one annotation
    for vn_sentence_idx, entry_list in parallel_VN2entries_for_dataset_dict.items():
        #print("vn_sentence_idx={}, entry_list with {} entries: {}".format(vn_sentence_idx, len(entry_list), entry_list))
        needed_pred_idx_list = [x[utils.MAPPING_FILE.SR3deVN_tokidx.value] for x in entry_list]
        
        #salsa annotation
        salsa_sentenceid = entry_list[entry_index][utils.MAPPING_FILE.SALSA_sentid.value] # take one sentence once
        salsa_frames_dict = salsa_sentenceid_to_frames_dict[salsa_sentenceid]
        #print("")
        filtered_salsa_frames_dict = __filter_salsa_sentenceid_to_frames_dict(salsa_frames_dict, needed_pred_idx_list)
        
        fully_annotated_VN_sentence = fully_annotated_VN_sentence_list[sentence_idx_in_vn_list]
        
        #merge them
        salsa_annotated_sentence = __replace_conll_with_salsa_annotation(salsa_sentenceid, filtered_salsa_frames_dict, fully_annotated_VN_sentence)
        fully_annotated_FN_sentence_list.append(salsa_annotated_sentence)
        
        sentence_idx_in_vn_list += 1
        
    return fully_annotated_FN_sentence_list

def __filter_salsa_sentenceid_to_frames_dict(salsa_frames_dict, needed_pred_idx_list):
    filtered_salsa_frames_dict = salsa_frames_dict
    
    return filtered_salsa_frames_dict


def __replace_conll_with_salsa_annotation(salsa_sentenceid, salsa_frames_dict_for_sentence, conll2009_sentence):
    
    #print("___")
    #print("FOR salsa_sentenceid: ", salsa_sentenceid)
    #print("salsa_frames_dict_for_sentence:", salsa_frames_dict_for_sentence)
    ##[{(pred_idx, label): [list_of_args_with_labels]}]
    # arg = (token_idx, arg_label)
    conll_pas_annotations = CoNLL2009reader.get_pas_annotation_list(conll2009_sentence)
    #print("conll_pas_annotations: {}".format(conll_pas_annotations))
    
    pred_mapping_dict_list = []
    arg_mapping_dict_list = []
    for predicate_position in range(len(conll_pas_annotations)):
        #print("  predicate position: ", predicate_position)
        conll_pas_annotation = conll_pas_annotations[predicate_position]
        #print("  conll_pas_annotation: ", conll_pas_annotation)
        
        conll_pred_idx, conll_pred_label = list(conll_pas_annotation.keys())[0] #only one item
        conll_argidxlabel_list = list(conll_pas_annotation.values())[0]
        conll_argid_to_argidx_dict = {}
        for argidx, arglabel in conll_argidxlabel_list:
            conll_argid_to_argidx_dict[argidx+1] = argidx
        #print("  for current position: conll_pred_idx:{}, conll_pred_label:{}, conll_argid_to_argidx_dict:{}".format(conll_pred_idx, conll_pred_label, conll_argid_to_argidx_dict))
        # conll_tokidx : salsa_label
        pred_mapping_dict, arg_mapping_dict = __get_salsa_pred_and_arg_mapping_dicts(conll2009_sentence, salsa_frames_dict_for_sentence, salsa_sentenceid, 
    conll_pred_idx, conll_argid_to_argidx_dict)
        
        pred_mapping_dict_list.append(pred_mapping_dict)
        arg_mapping_dict_list.append(arg_mapping_dict)
        #print("    pred_mapping_dict_list:",pred_mapping_dict_list)
        #print("    arg_mapping_dict_list:",arg_mapping_dict_list)
    
    conll2009_sentence_toChange = __copy(conll2009_sentence)    
        
    #replace predicates and check them 
    changed_conll_predtokidx_list = []
    for pred_mapping_dict in pred_mapping_dict_list:
        for conll_predtokidx, salsa_predlabel in pred_mapping_dict.items():
            orig_conll_label = conll2009_sentence[conll_predtokidx][CoNLL2009reader.CONLL2009_ANNOIDX.PRED.value]
            if orig_conll_label == CoNLL2009reader.FORMAT.NOVALUE.value:
                raise IndexError("MISTAKEN (FP) PREDICATE at conll predtoken idx={} for SALSA label={}".format(conll_predtokidx, salsa_predlabel))
            
            #print("-> changed predicate: conllidx={}, conlllabel={} => salsalabel={}".format(conll_predtokidx, orig_conll_label, salsa_predlabel))
            conll2009_sentence_toChange[conll_predtokidx][CoNLL2009reader.CONLL2009_ANNOIDX.PRED.value] = salsa_predlabel
            changed_conll_predtokidx_list.append(conll_predtokidx)
    
    #check predicates
    for conll_idx in range(len(conll2009_sentence)):
        orig_conll_pred_label = conll2009_sentence[conll_idx][CoNLL2009reader.CONLL2009_ANNOIDX.PRED.value]
        
        if orig_conll_pred_label == CoNLL2009reader.FORMAT.NOVALUE.value:
            if conll_idx in changed_conll_predtokidx_list:
                raise IndexError("FALSE POSITIVE predicate with idx={} in sentence \n{}".format(conll_idx, conll2009_sentence))
        else:
            if conll_idx not in changed_conll_predtokidx_list:
                msg = "FALSE NEGATIVE predicate with idx={} (is labelled with conll:{} ) in sentence \n{}".format(conll_idx, orig_conll_pred_label, conll2009_sentence)
                raise IndexError(msg)
        
    #replace arguments and check them
    for conll_predposition in range(len(arg_mapping_dict_list)):
        changed_argtokidx_list = []
        
        #replace
        arg_column_idx = CoNLL2009reader.CONLL2009_ANNOIDX.APRED1.value + conll_predposition
        
        for conll_argtokidx, salsa_arglabel in arg_mapping_dict_list[conll_predposition].items():
            #print("arg_mapping_dict for conll_predpostion={}: {}".format(conll_predposition, arg_mapping_dict))
            orig_conll_label = conll2009_sentence[conll_argtokidx][arg_column_idx]
            #print("--> changed argument: conlltokidx={}, conllcolumnidx={}, conlllabel={} => salsalabel={}".format(conll_argtokidx, arg_column_idx, orig_conll_label, salsa_arglabel))
            conll2009_sentence_toChange[conll_argtokidx][arg_column_idx] = salsa_arglabel
            changed_argtokidx_list.append(conll_argtokidx)
         
        #check arguments
        for conll_idx in range(len(conll2009_sentence)):
            orig_conll_arg_label = conll2009_sentence[conll_idx][arg_column_idx]
            
            if orig_conll_arg_label == CoNLL2009reader.FORMAT.NOVALUE.value:
                if conll_idx in changed_argtokidx_list:
                    raise IndexError("FALSE POSITIVE argument with idx={}, columnidx={}, in sentence \n{}".format(conll_idx, arg_column_idx, conll2009_sentence))
            else:
                if conll_idx not in changed_argtokidx_list:
                    #remove the conll argument
                    conll2009_sentence_toChange[conll_idx][arg_column_idx] = CoNLL2009reader.FORMAT.NOVALUE.value
                    
                    #msg = "FALSE NEGATIVE argument with idx={}, columnidx={}, orig_conll_label:{} in sentence \n{}".format(conll_idx, arg_column_idx, orig_conll_arg_label, conll2009_sentence)
                    #raise IndexError(msg)
                    
                    conll_pred_dict = pred_mapping_dict_list[conll_predposition]
                    conll_pred_idx, salsa_pred_label = list(conll_pred_dict.items())[0]
#maybe interesting                    msg = "  Dropped SALSA frame element (for conll argument '{}', for the {}. predicate with sense '{}') annotated in the corresponding CoNLL2009 sentence\n\t'{}'\n".format(orig_conll_arg_label, conll_predposition+1, conll2009_sentence[conll_pred_idx][CoNLL2009reader.CONLL2009_ANNOIDX.PRED.value], " ".join([str(x[CoNLL2009reader.CONLL2009_ANNOIDX.FORM.value]) for x in conll2009_sentence]))
#                    print(msg) # make only a message
                    
                    #print("original_conll_sentence:\n{}\nsalsa_sentence:\n{}\n".format("\n".join([str(x) for x in conll2009_sentence]), "\n".join([str(x) for x in conll2009_sentence_toChange])))
    return conll2009_sentence_toChange


def __get_salsa_pred_and_arg_mapping_dicts(conll2009_sentence, salsa_frames_dict_for_sentence, salsa_sentenceid, 
    conll_pred_idx, conll_argid_to_argidx_dict):
    #for one predicate
    pred_mapping_dict = {} #with only one predicate
    arg_mapping_dict = {}
    
    #print("salsa_frames_dict_for_sentence: ", salsa_frames_dict_for_sentence)
    #print("type(salsa_frames_dict_for_sentence): ", type(salsa_frames_dict_for_sentence))
    
    appropriate_fe_content_dict = None
    salsa_target_alternatives = {} #{ conll_pred_idx: [ fe_content_dict,... ] }
    for salsa_frame_wholeid, frame_content in salsa_frames_dict_for_sentence.items():
        found_pred = False
        #print("salsa_frame_wholeid:{}, frame_content:{}".format(salsa_frame_wholeid, frame_content))
        #search for the given annotation
        
        #frame content: [
        #       { target_name : [(sentid, [list of tids]), (sentid, [list of tids]), ...] }
        #       { fe_name : [(sentid, [list of tids]), ...], fe_name : [...], ... }
        # ]
        #print("frame_content for salsa_frame_id {} : {}".format(salsa_frame_wholeid, frame_content))
        target_content_dict, fe_content_dict = frame_content

        salsa_target_name = list(target_content_dict.keys())[0] #Assumption: only one annotation is there
        salsa_target_tokens = list(target_content_dict.values())[0]
        
        for target_sentid, target_tokid_list in salsa_target_tokens:
            if target_sentid != salsa_sentenceid: #target beyond the sentence (not assumed)
                continue
            
            for target_tokid in target_tokid_list:
                if conll_pred_idx + 1 == target_tokid:
                    #found predicate
                    pred_mapping_dict[conll_pred_idx] = salsa_target_name #entails always the last one
                    appropriate_fe_content_dict = fe_content_dict
                    salsa_target_alternatives.setdefault(conll_pred_idx, [])
                    salsa_target_alternatives[conll_pred_idx].append(fe_content_dict)
                    found_pred = True
                    break
            #if found_pred == True: #NOTE do not break here !
            #    break
        
    if appropriate_fe_content_dict is None:
        raise ValueError("NOT FOUND SALSA TARGET in CONLL sentence with proposed predicate line idx:{} -- conll sentence:\n{}\n\nsalsa_frames_dict_for_sentence: {}".format(conll_pred_idx, conll2009_sentence, salsa_frames_dict_for_sentence))
        #search for arguments
    else:
        salsa_target_alternatives[conll_pred_idx].remove(appropriate_fe_content_dict)
        
        for salsa_arg_id, conll_arg_idx in conll_argid_to_argidx_dict.items():
            found_arg = False
            for fe_name, fe_tokens in appropriate_fe_content_dict.items():
                for fe_sentid, fe_tokid_list in fe_tokens:
                    if fe_sentid is None or fe_sentid != salsa_sentenceid: #fe beyond the sentence 
                        continue
                    for fe_tokid in fe_tokid_list:
                        if fe_tokid == salsa_arg_id:
                            arg_mapping_dict[conll_arg_idx] = fe_name
                            found_arg = True
                            break
                    if found_arg == True:
                        break
                if found_arg == True:
                    break
            if found_arg == False:
                if len(salsa_target_alternatives[conll_pred_idx]) > 0:
                    #one more option
                    arg_mapping_dict = {}
                    alterntive_fe_content_dict = salsa_target_alternatives[conll_pred_idx][0]
                    #print("ALTERNATIVE FE CONTENT DICT for conll_pred_idx: {} = {}".format(conll_pred_idx, alterntive_fe_content_dict))
                    for salsa_arg_id, conll_arg_idx in conll_argid_to_argidx_dict.items():
                        found_arg = False
                        for fe_name, fe_tokens in alterntive_fe_content_dict.items():
                            for fe_sentid, fe_tokid_list in fe_tokens:
                                if fe_sentid is None or fe_sentid != salsa_sentenceid: #fe beyond the sentence 
                                    continue
                                for fe_tokid in fe_tokid_list:
                                    if fe_tokid == salsa_arg_id:
                                        arg_mapping_dict[conll_arg_idx] = fe_name
                                        found_arg = True
                                        break
                                if found_arg == True:
                                    break
                            if found_arg == True:
                                break
                        if found_arg == False:
                            #FALSE NEGATIVE ARGUMENT --> will be prompted later
                            pass
                            #msg = "NOT FOUND SALSA FE WITH ALTERNATIVE FE CONTENT for conll arg line idx:'{}' in CONLL sentence with proposed predicate line idx:{}\nfound salsa correspondences: arg: {}; pred:{}\n -- conll sentence:\n{}\n\nsalsa_frames_dict_for_sentence: \n{}".format(conll_arg_idx, conll_pred_idx, arg_mapping_dict, pred_mapping_dict, "\n".join([str(x) for x in conll2009_sentence]), salsa_frames_dict_for_sentence)
                            #raise IndexError(msg)
                            
                else:
                    #FALSE NEGATIVE ARGUMENT --> will be prompted later
                    pass
                    #msg = "NOT FOUND SALSA FE (no alternative there) for conll arg line idx:'{}' in CONLL sentence with proposed predicate line idx:{}\nfound salsa correspondences: arg: {}; pred:{}\n -- conll sentence:\n{}\n\nsalsa_frames_dict_for_sentence: \n{}".format(conll_arg_idx, conll_pred_idx, arg_mapping_dict, pred_mapping_dict, "\n".join([str(x) for x in conll2009_sentence]), salsa_frames_dict_for_sentence)
                    #raise IndexError(msg)
                    
        
    return (pred_mapping_dict, arg_mapping_dict)


def __copy(conll2009_sentence):
    copy_list_of_list = []
    for line_split in conll2009_sentence:
        copy_list = []
        for item in line_split:
            copy_list.append(item)
        copy_list_of_list.append(copy_list)
    return copy_list_of_list
        
    
def __cut_off_srl_annotation(conll_formatted_sentence):
    conll_base_sentence = []
    
    for conll_formatted_line in conll_formatted_sentence:
        conll_base_line = conll_formatted_line[:CoNLL2009reader.CONLL2009_ANNOIDX.FILLPRED.value]
        conll_base_sentence.append(conll_base_line)
    
    return conll_base_sentence
    
